/* This file is a part of TSP Flaming
 *  Copyright (C) 2008-2010 Nikolay Kurtov <nkurtov@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */
 
/* Author: Nikolay Kurtov <nkurtov@gmail.com>
 *      ___            __                   __
 *     / __`\         /  _\                /\ \ 
 *    /\ \/\ \  _ __ /\ \_/    ___     __  \_\ \__
 *    \ \ \ \ \/\`'__\_\ \__  /'__`\ /`___\/\__  _\ 
 *     \ \ \_\ \ \ \//\__  _\/\  __//\____ \/_/\ \/_
 *      \ \____/\ \_\\/_/\ \/\ \____\/_____/  \ \___\ 
 *       \/___/  \/_/   \_\ \ \/____/_____/    \/__/
 *                      /\__/
 *                      \/_/
 */        

#include "calculator.h"

#include <QMutex>

#include <algorithm>
#include <math.h>

Calculator::Calculator() : running(false), iteration(0), temperature(0), steps(0), cooling(0), freezes(0), 
				currentSolution(NULL), bestSolution(NULL), resCount(0), resPoints(new double[Calculator::MAX_POINTS]),
				mutex(new QMutex()){
}

Calculator::~Calculator(){
	delete[] this->resPoints;
	delete[] this->currentSolution;
	delete[] this->bestSolution;
}

void Calculator::setPoints(const QVector<Point>& aPoints){
	Q_ASSERT(this->running == false);
	this->points = aPoints;
	if (this->currentSolution){
		delete[] this->currentSolution;
		delete[] this->bestSolution;
	}
	this->currentSolution = new int[this->points.size()];
	this->bestSolution = new int[this->points.size()];
	for (int index = 0; index < this->points.size(); index++){
		this->currentSolution[index] = index;
		this->bestSolution[index] = index;
	}
}

void Calculator::runRecved(double aTemperature, int aSteps, double aCooling, int aFreezes){
	Q_ASSERT(this->running == false);
	this->running = true;
	this->iteration = 0;
	this->temperature = aTemperature;
	this->steps = aSteps;
	this->cooling = aCooling;
	this->freezes = aFreezes;
	this->resCount = 0;
	this->start();
}

void Calculator::stopRecved(){
	Q_ASSERT(this->running == true);
	this->running = false;
}

void Calculator::run(){
	Q_ASSERT(this->running);
	for (int index = 0; index < this->points.size(); index++) {
		this->bestSolution[index] = this->currentSolution[index];
	}
	this->calculateLength();
	this->bestResult = this->currentResult;
	this->iteration++;
	int* oldPoints = new int[this->points.size()];
	while (this->running && this->freezes > 0) {
		bool better = false;
		for (int index = 0; index < this->steps && this->running; index++) {
			std::copy(this->currentSolution, this->currentSolution + this->points.size(), oldPoints);
			double oldResult = this->currentResult;
                        this->mutex->lock();
			this->getNextSolution();
                        this->mutex->unlock();
			double diff = this->currentResult - oldResult;
			if (diff < 0) {
				if (this->currentResult < this->bestResult) {
					this->bestResult = this->currentResult;
					std::copy(this->currentSolution, this->currentSolution + this->points.size(), this->bestSolution);
					better = true;
				}
			} else {
				double rndm = qrand() / double (RAND_MAX);
				double div = -diff / double (this->temperature);
				double pow = exp (div);
				if (rndm >= pow) {
					this->mutex->lock();
					std::copy(oldPoints, oldPoints + this->points.size(), this->currentSolution);
					this->mutex->unlock();
					this->currentResult = oldResult;
				}
			}
			this->iteration++;
			if (this->resCount > Calculator::MAX_POINTS - 1) {
				this->resCount = 0;
			}
			this->resPoints[this->resCount] = this->currentResult;
			resCount++;
		}
		this->temperature *= this->cooling;
		if (!better) {
			this->freezes--;
		}
	}
	running = false;
	delete[] oldPoints;
	std::copy(this->currentSolution, this->currentSolution + this->points.size(), this->bestSolution);
}

void Calculator::calculateLength(){
	double value = 0;
	int size = this->points.size();
	for (int index = 0; index < size - 1; index++){
		value += dist(this->points[currentSolution[index+1]], this->points[currentSolution[index]]);
	}
	value += dist(this->points[currentSolution[size-1]], this->points[currentSolution[0]]);
	this->currentResult = value;
}

void Calculator::getNextSolution(){
	if (int (6.0 * (::qrand() / (RAND_MAX + 1.0))) == 0){
		this->getSolutionAType();
	} else {
		this->getSolutionBType();
		this->calculateLength();
	}
}

void Calculator::getSolutionAType(){
	int first;
	int second;
	int diff = 0;
	while (diff < 2){
		first = ::qrand() % this->points.size();
		second = ::qrand() % this->points.size();
		if (first > second){
			first ^= second ^= first ^= second;
		}
		Q_ASSERT(first <= second);
		diff = second - first;
	}
	int n_first = first+1;
	int n_second = (second+1) % this->points.size();
	double old = (this->points[this->currentSolution[n_first]] - this->points[this->currentSolution[first]]).length() + 
				(this->points[this->currentSolution[n_second]] - this->points[this->currentSolution[second]]).length();
	std::reverse (this->currentSolution + first + 1, this->currentSolution + second + 1);
	double add = (this->points[this->currentSolution[n_first]] - this->points[this->currentSolution[first]]).length() + 
				(this->points[this->currentSolution[n_second]] - this->points[this->currentSolution[second]]).length();
				
	this->currentResult += add - old;
}

void Calculator::getSolutionBType(){
	int first;
	int second;
	int third;
	bool good = false;
	while (!good) {
		first = int (this->points.size() * (::qrand() / (RAND_MAX + 1.0)));
		second = int (this->points.size() * (::qrand() / (RAND_MAX + 1.0)));
		third = int (this->points.size() * (::qrand() / (RAND_MAX + 1.0)));
		if (first > second) {
			first ^= second ^= first ^= second;
		}
		if (first > third) {
			first ^= third ^= first ^= third;
		}
		if (second > third) {
			third ^= second ^= third ^= second;
		}
		Q_ASSERT(first <= second && second <= third);
		if (third - second < 2 || second - first < 2) {
			continue;
		}
		good = true;
	}
	std::rotate(this->currentSolution, this->currentSolution + first + 1, this->currentSolution + this->points.size());
	int old_second = second;
	second = third - first - 1;
	first = old_second - first - 1;
	int strategy = int (5.0 * (::qrand() / (RAND_MAX + 1.0)));
	if (strategy == 0) {	// a r_b r_c
		std::reverse (this->currentSolution + first + 1, this->currentSolution + second + 1);
		std::reverse (this->currentSolution + second + 1, this->currentSolution + this->points.size());
	} else if (strategy == 1) {	// a c r_b
		std::reverse (this->currentSolution + first + 1, this->currentSolution + second + 1);
		this->mix (first, second);
	} else if (strategy == 2) {	// a c b
		this->mix (first, second);
	} else if (strategy == 3) {	// a r_c r_b
		std::reverse (this->currentSolution + first + 1, this->currentSolution + second + 1);
		std::reverse (this->currentSolution + second + 1, this->currentSolution + this->points.size());
		this->mix (first, second);
	} else if (strategy == 4) {	// a r_c b
		std::reverse (this->currentSolution + second + 1, this->currentSolution + this->points.size());
		this->mix (first, second);
	}
}

void Calculator::mix (int first, int second) {
	int m_first = first;
	int m_second = first + this->points.size() - second - 1;
	int* tmp = new int[this->points.size()];
	std::copy (this->currentSolution, this->currentSolution + this->points.size(), tmp);
	std::copy (tmp + first + 1, tmp + second + 1, this->currentSolution + m_second + 1);
	std::copy (tmp + second + 1, tmp + this->points.size(), this->currentSolution + m_first + 1);
	delete[] tmp;
}
