/* This file is a part of TSP Flaming
 *  Copyright (C) 2008-2010 Nikolay Kurtov <nkurtov@gmail.com>
 *
 * Authors:
 *  Nikolay Kurtov <nkurtov@gmail.com>
 *  Serguey Basalaev <sbasalaev@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

/* 
 *      ___            __                   __
 *     / __`\         /  _\                /\ \ 
 *    /\ \/\ \  _ __ /\ \_/    ___     __  \_\ \__
 *    \ \ \ \ \/\`'__\_\ \__  /'__`\ /`___\/\__  _\ 
 *     \ \ \_\ \ \ \//\__  _\/\  __//\____ \/_/\ \/_
 *      \ \____/\ \_\\/_/\ \/\ \____\/_____/  \ \___\ 
 *       \/___/  \/_/   \_\ \ \/____/_____/    \/__/
 *                      /\__/
 *                      \/_/
 */        

#include "mainWindow.h"

#include <QDesktopServices>
#include <QFileDialog>
#include <QMessageBox>
#include <QTextStream>
#include <QTimer>
#include <QString>
#include <QUrl>

#include "point.h"

/** Program name. */
const QLatin1String MainWindow::NAME("TSP Flaming");
/** Program version. */
const QLatin1String MainWindow::VERSION("v1.2");

/**
 * Creates <c>MainWindow</c> object.
 * If <c>filename</c> is specified then file is loaded.
 * @param filename a name of the file to load or empty string
 */
MainWindow::MainWindow(const QString& filename = QString()) :
		running(false), calculator(new Calculator()), timer(new QTimer()), task(new Task()), oldSteps(1){
	this->setupUi(this);
	this->paintArea->setPoints(this->calculator->getResPoints());
	this->mapArea->setMutex(this->calculator->getMutex());
	this->mapArea->setSolutionPtr(this->calculator->getCurrentSolutionPtr());
	
	connect(this->actionOpen, SIGNAL (activated()), this, SLOT (open()));
	connect(this->actionSave, SIGNAL (activated()), this, SLOT (save()));
	
	connect(this, SIGNAL(runCalculator(double, int, double, int)), this->calculator, SLOT(runRecved(double, int, double, int)));
	connect(this, SIGNAL(stopCalculator()), this->calculator, SLOT(stopRecved()));
	connect(this->calculator, SIGNAL(finished()), this, SLOT(calculatorStopped()));

	connect(this->mapTable, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(verifyCell(QTableWidgetItem*)));
	
	this->statusBar()->showMessage(tr("Ready"), STATUS_BAR_DELAY);
	if (!filename.isEmpty()) this->loadFile(filename);
}

/**
 * Shows file open dialog and opens the selected file.
 */
void MainWindow::open(){
	if (this->running == true){
		this->triggerCalculations();
	}
	QString fileName = QFileDialog::getOpenFileName(
		this,
		tr("Open TSP task"),
		QString(),
		tr("TSP tasks (*.tsptask);;All files (*)"));
	if (fileName.isEmpty()) return;
	if (this->loadFile (fileName)){
		this->statusBar()->showMessage(tr("Task loaded"), STATUS_BAR_DELAY);
		this->setWindowTitle(fileName.mid(fileName.lastIndexOf('/')+1) + " - " + NAME);
	} else {
		this->statusBar()->showMessage(tr("Can not load the file"), STATUS_BAR_DELAY);
	}
}

/**
 * Loads TSP task from the specified file.
 *
 * @param fileName the name of the file
 * @return true if file was successfully loaded, false otherwise
 */
bool MainWindow::loadFile(const QString& fileName){
	if (!this->task->load (fileName)) return false;
	this->nameEdit->setText(this->task->name());
	this->commentEdit->setPlainText(this->task->comments());
	this->calculator->setPoints(task->points());
	this->mapArea->setPoints(task->points());
	this->mapArea->update();
	this->updateMapTable();
	return true;
}

void MainWindow::triggerCalculations(){
	if (this->running == false){
		if (this->calculator->getCurrentSolution() == NULL){
			QMessageBox::warning(this, NAME, tr("Task is not loaded"));
			return;
		}
		this->runButton->setText(tr("Stop"));
		this->paintArea->reset();
		
		double initTemperature = this->tempEdit->text().toDouble();
		int initSteps = this->stepsEdit->text().toInt();
		this->oldSteps = initSteps;
		double initCooling = this->coolingEdit->text().toDouble();
		int initFreezes = this->freezesEdit->text().toInt();

		emit runCalculator(initTemperature, initSteps, initCooling, initFreezes);
		
		this->statusBar()->showMessage(tr("Calculation started"), STATUS_BAR_DELAY);
		this->running = true;
		this->editTab->setEnabled(false);
		connect(this->timer, SIGNAL(timeout()), this, SLOT(updateData()));
		this->timer->start(100);
	} else {
		emit stopCalculator();
		this->calculatorStopped();
	}
}

void MainWindow::calculatorStopped(){
	this->updateData();
	this->running = false;
	this->runButton->setText(tr("Start"));
	this->editTab->setEnabled(true);
	this->statusBar()->showMessage(tr("Calculation stopped"), STATUS_BAR_DELAY);
	disconnect(this->timer, SIGNAL(timeout()), this, SLOT(updateData()));
}

/**
 * Opens save file dialog and saves task into the selected file.
 */
void MainWindow::save(){
	if (this->running == true){
		this->triggerCalculations();
	}
	QString fileName = QFileDialog::getSaveFileName(
		this,
		tr("Save TSP task"),
		QString(),
		tr("TSP tasks (*.tsptask);;All files (*)"));
	if (!fileName.isEmpty()) {
		if (this->saveFile (fileName)){
			this->statusBar()->showMessage(tr("Task saved"), STATUS_BAR_DELAY);
		}
	}
}

/**
 * Saves task into the specified file.
 * @param fileName a name of the file
 */
bool MainWindow::saveFile(const QString& fileName){
	this->task->setName(this->nameEdit->text());
	this->task->setComments(this->commentEdit->toPlainText());
	return this->task->save(fileName);
}

/* Original saveFile() which in fact saves solution. */
//bool MainWindow::saveFile(const QString& fileName){
//	const QVector<Point>& points = this->calculator->getPoints();
//	const int* const indices = this->calculator->getBestSolution();
//	int size = points.size();
//	QFile file (fileName);
//	if (!file.open (QFile::WriteOnly | QFile::Text)) {
//		QMessageBox::warning (this, tr ("TSP Flaming"),
//		                      tr ("Cannot write to file %1:\n%2.").arg (fileName).arg (file.errorString()));
//		return false;
//	}
//	QTextStream out (&file);
//	for (int index = 0; index < size; index++) {
//		out << indices[index] << " " << points[indices[index]].getX() << " " << points[indices[index]].getY() << "\n";
//	}
//	file.close();
//	return true;
//}

void MainWindow::updateData(){
	this->paintArea->setPointsNumber(this->calculator->getResCount());
	this->paintArea->update();
	
	this->mapArea->update();
	
	int iteration = this->calculator->getIteration();
	QString iterationValueString = QString::number(iteration);
	
	double currentResult = this->calculator->getCurrentResult();
	QString currentResultString = QString::number(currentResult);
	double bestResult = this->calculator->getBestResult();
	QString bestResultString = QString::number(bestResult);
	double currentTemperature = this->calculator->getTemperature();
	QString currentTemperatureString = QString::number(currentTemperature);
	double freezes = this->calculator->getFreezes();
	QString freezesString = QString::number(freezes);
	
	this->iterValue->setText(iterationValueString);
	this->currentSolutionValue->setText(currentResultString);
	this->bestSolutionValue->setText(bestResultString);
	this->temperatureValue->setText(currentTemperatureString);
	this->freezesValue->setText(freezesString);
	
	int oldfreezes = this->freezesEdit->text().toInt();
	this->progressBar->setValue(100*(oldfreezes - freezes)/oldfreezes);
	
	timer->start (0);
}

/** Updates map table from the task data. */
void MainWindow::updateMapTable(){
	mapTable->setRowCount(task->dimension());
	for (int i=0; i<task->dimension(); i++){
		mapTable->setItem(i, 0, new QTableWidgetItem(QString::number(task->points()[i].getX())));
		mapTable->setItem(i, 1, new QTableWidgetItem(QString::number(task->points()[i].getY())));
	}
}

/**
 * Removes selected point from the task and
 * from the <c>mapTable</c>.
 */
void MainWindow::removeTaskPoint(){
	//removing table row
	disconnect(this->mapTable, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(verifyCell(QTableWidgetItem*)));
	int row = mapTable->currentRow();
	if (row == -1) return;
	QTableWidgetItem* item = mapTable->takeItem(row, 0);
	delete item;
	item = mapTable->takeItem(row, 1);
	delete item;
	mapTable->removeRow(row);
	connect(this->mapTable, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(verifyCell(QTableWidgetItem*)));
	//removing a point from the task
	task->removePoint(row);
}

/**
 * Adds new point to the task and to the <c>mapTable</c>.
 */
void MainWindow::addTaskPoint(){
	this->addTaskPoint(Point(0,0));
}

/**
 * Adds new point to the task and to the <c>mapTable</c>.
 * @param a point to add
 */
void MainWindow::addTaskPoint(const Point& point){
	//adding a point to the task
	task->addPoint(Point(0,0));
	//adding table row
	disconnect(this->mapTable, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(verifyCell(QTableWidgetItem*)));
	mapTable->setRowCount(mapTable->rowCount()+1);
	mapTable->setItem(mapTable->rowCount()-1, 0, new QTableWidgetItem(QString::number(point.getX())));
	mapTable->setItem(mapTable->rowCount()-1, 1, new QTableWidgetItem(QString::number(point.getY())));
	connect(this->mapTable, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(verifyCell(QTableWidgetItem*)));
}

/**
 * @brief Verifies cell value and updates task.
 * If value in the cell is a valid floating point value
 * then task is updated to have that value, otherwise
 * cell is shown in red.
 * @param item item in table that has been changed
 */
void MainWindow::verifyCell(QTableWidgetItem* item){
	bool ok;
	double newValue = item->text().toDouble(&ok);
	if (ok) {
		Point oldpoint = task->points()[item->row()];
		if (item->column() == 0)
			task->modifyPoint(item->row(), Point(newValue, oldpoint.getY()));
		else
			task->modifyPoint(item->row(), Point(oldpoint.getX(), newValue));
		item->setBackground(QBrush(Qt::white));
	} else {
		item->setBackground(QBrush(Qt::red));
	}
}

/** Shows information about Qt. */
void MainWindow::about(){
	QMessageBox msgBox (this);
	msgBox.setIconPixmap(this->windowIcon().pixmap(64,64));
	msgBox.setText("<h1>" +NAME + ' ' + VERSION + "</h1>");
	msgBox.setInformativeText( QString() +
		"<p>Copyright (c) 2008-2010, Nikolay Kurtov</p>" +
		"<p><b>" + tr("Developers:")+ "</b>" +
		Qt::convertFromPlainText(
			"Nikolay Kurtov <nkurtov@gmail.com>\nSerguey Basalaev <sbasalaev@gmail.com>"
		) +
		"</p><p><b>" +
		tr("Translators:") + "</b>" +
		Qt::convertFromPlainText(tr("translator-credits")) +
		"</p><p><b>" + tr("License:") + "</b></p>" +
		"<p>This program is free software: you can redistribute it and/or modify " +
		"it under the terms of the GNU General Public License as published by " +
		"the Free Software Foundation, either version 3 of the License, or " +
		"any later version.</p>" +
		"<p>This program is distributed in the hope that it will be useful, " +
		"but WITHOUT ANY WARRANTY; without even the implied warranty of " +
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the " +
		"GNU General Public License for more details.</p>" +
		"<p>You should have received a copy of the GNU General Public License " +
		"along with this program.  If not, see " +
		"<a href=\"http://www.gnu.org/licenses/\">http://www.gnu.org/licenses/</a>.</p>"
	);
	msgBox.setStandardButtons(QMessageBox::Ok);
	msgBox.setDefaultButton(QMessageBox::Ok);
	msgBox.exec();
}

/** Shows information about Qt. */
void MainWindow::aboutQt(){
	QMessageBox::aboutQt(this);
}

/** Shows online help. */
void MainWindow::help(){
	QUrl url = tr("http://tspflaming.sourceforge.net");
	QDesktopServices::openUrl(url);
}
