/* This file is a part of TSP Flaming
 *  Author: Serguey Basalaev <sbasalaev@gmail.com>
 *  Copyright (C) 2008-2010 Nikolay Kurtov <nkurtov@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include <QTextStream>
#include <QMessageBox>
#include "task.h"
#include "point.h"

/**
 * Returns the name of this task.
 * @return task name
 */
QString Task::name() const{
	return this->fName;
}

/**
 * Returns the comments to the task.
 * @return task comments
 */
QString Task::comments() const{
	return this->fComments;
}

/**
 * Returns the dimension of the task.
 * @return task dimension
 */
int Task::dimension() const{
	//FIXME: return dimension which is read from the file
	//DEPENDS: calculator which respects this getter
	return this->fPoints.size();
}

/**
 * Returns a vector with the task points.
 * @return task points
 */
const QVector<Point>& Task::points() const{
	return this->fPoints;
}

/**
 * Sets new name to the task.
 * @param name new task name
 */
void Task::setName(const QString& name){
	this->fName = name;
}

/**
 * Sets new comments to the task.
 * @param comments new task comments
 */
void Task::setComments(const QString& comments){
	this->fComments = comments;
}

/**
 * Adds point to the task, increasing it's dimension.
 * @param point new point
 */
void Task::addPoint(const Point& point){
	this->fPoints.append(point);
}

/**
 * Modifies specified point of the task.
 * @param pos position of point to modify
 * @param point new point
 */
void Task::modifyPoint(int pos, const Point& point){
	this->fPoints[pos] = point;
}

/**
 * Removes specified point from the task decreasing it's dimension.
 * @param pos position of the point to remove
 */
void Task::removePoint(int pos){
	this->fPoints.remove(pos);
}

/**
 * Loads task from the specified file.
 * If file is not a valid TSP task then the function
 * shows error messages to the user.
 * @param fileName the name of the file to load from
 * @return <c>true</c> if task was successfully loaded,
 *         <c>false</c> otherwise
 */
bool Task::load(const QString& fileName){
	//reading file in format introduced by Nikolay Kurtov
	QFile file (fileName);
	QString newName;
	QString newComments;
	QVector<Point> newPoints;
	if (!file.open (QFile::ReadOnly | QFile::Text)){
		QMessageBox::warning (NULL, tr("I/O error"),
			tr ("Cannot read file %1:\n%2.").arg (file.fileName(), file.errorString()));
		return false;
	}
	bool nameRead = false, commentRead = false, dimRead =  false,
	     typeRead = false, pointsRead = false;
	QString ERROR = tr("Parsing error");
	QString ERRMESSAGE = tr("In file %1:\nAt line %2:\n%3").arg(fileName);
	QTextStream in (&file);
	QString line = in.readLine();
	int lineNumber = 1;
	while (!in.atEnd()){
		if (line.startsWith(QLatin1String("NAME:"))){
			if (nameRead) QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Multiple \"%1\" invocations").arg("NAME")));
			newName = line.mid(5).trimmed();
			nameRead = true;
		} else if (line.startsWith(QLatin1String("COMMENT:"))){
			if (commentRead) QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Multiple \"%1\" invocations").arg("COMMENT")));
			newComments = line.mid(8).trimmed();
			line = in.readLine();
			lineNumber++;
			while (!line.isNull() && line.startsWith(QLatin1Char(' '))) {
				newComments.append('\n').append(line.trimmed());
				line = in.readLine();
			}
			commentRead = true;
			continue;
		} else if (line.startsWith(QLatin1String("EDGE_WEIGHT_TYPE:"))) {
			if (typeRead) QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Multiple \"%1\" invocations").arg("EDGE_WEIGHT_TYPE")));
			QString type = line.mid(17).trimmed();
			if (type != QLatin1String("EUC_2D")) {
				QMessageBox::critical(NULL, ERROR,
					ERRMESSAGE.arg(lineNumber).arg(tr("Unsupported type %1").arg(type)));
				return false;
			}
			typeRead = true;
		} else if (line.startsWith(QLatin1String("DIMENSION:"))) {
			if (dimRead) QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Multiple \"%1\" invocations").arg("DIMENSION")));
			bool ok;
			QString dim = line.mid(10).trimmed();
			this->fDimension = dim.toInt(&ok);
			if (this->fDimension < 2) ok = false;
			if (!ok) {
				QMessageBox::critical(NULL, ERROR,
					ERRMESSAGE.arg(lineNumber).arg(tr("Wrong dimension of task: %1").arg(dim)));
				return false;
			}
			dimRead = true;
		} else if (line.startsWith(QLatin1String("NODE_COORD_SECTION"))){
			if (pointsRead) QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Multiple \"%1\" invocations").arg("NODE_COORD_SECTION")));
			if (!dimRead) {
				QMessageBox::critical(NULL, ERROR,
					ERRMESSAGE.arg(lineNumber).arg(tr("Dimension of task is not specified.")));
				return false;
			}
			if (!typeRead) {
				QMessageBox::critical(NULL, ERROR,
					ERRMESSAGE.arg(lineNumber).arg(tr("Type of task nodes is not specified.")));
				return false;
			}
			newPoints.clear();
			line = in.readLine();
			lineNumber++;
			while (!line.isNull() && !line.startsWith(QLatin1String("EOF"))){
				QStringList coordList = line.split(QRegExp(QLatin1String("\\s+")));
				if (coordList.size() < 3) {
					QMessageBox::critical(NULL, ERROR,
						ERRMESSAGE.arg(lineNumber).arg(tr("Insufficient data in the string \"%1\".").arg(line)));
					return false;
				}
				//don't really care for the coordList[0] to be increasing numbers
				//I think we might introduce labels instead in the future
				bool ok;
				double x = coordList[1].toDouble(&ok);
				if (!ok){
					QMessageBox::critical(NULL, ERROR,
						ERRMESSAGE.arg(lineNumber).arg(tr("Cannot convert %1 to double.").arg(coordList[1])));
					return false;
				}
				double y = coordList[2].toDouble(&ok);
				if (!ok){
					QMessageBox::critical(NULL, ERROR,
						ERRMESSAGE.arg(lineNumber).arg(tr("Cannot convert %1 to double.").arg(coordList[2])));
					return false;
				}
				if (coordList.size() > 3) {
					QMessageBox::warning(NULL, ERROR,
						ERRMESSAGE.arg(lineNumber).arg(tr("Trailing data in the string \"%1\".").arg(line)));
				}
                                newPoints.append(Point(x,y));
				line = in.readLine();
			}
                        if (newPoints.size() < fDimension) {
				QMessageBox::critical(NULL, ERROR,
					ERRMESSAGE.arg(lineNumber).arg(tr("The size of the node data section is less than task dimension.")));
				return false;
			}
			pointsRead = true;
		} else {
			QMessageBox::warning(NULL, ERROR,
				ERRMESSAGE.arg(lineNumber).arg(tr("Line \"%1\" is not recognized. Skipping.").arg(line)));
		}
		line = in.readLine();
		lineNumber++;
	}
	if (!(dimRead && pointsRead)) {
		QMessageBox::critical(NULL, ERROR,
			ERRMESSAGE.arg(lineNumber).arg(tr("File ended unexpectedly.")));
		return false;
	}
	this->setName(newName);
	this->setComments(newComments);
	this->fPoints.clear();
	this->fPoints += newPoints;
	return true;
}

/**
 * Saves task to the specified file.
 * @param fileName the name of the file to save to
 * @return <c>true</c> if file was saved successfully,
 *         <c>false</c> otherwise 
 */
bool Task::save(const QString& fileName){
	//saving to format introduced by Nikolay Kurtov
	QFile file (fileName);
	if (!file.open (QFile::WriteOnly | QFile::Text)) {
		QMessageBox::warning (NULL, tr("I/O error"),
			tr("Cannot write to file %1:\n%2.").arg (file.fileName()).arg(file.errorString()));
		return false;
	}
	QTextStream out (&file);
	out << "NAME: " << fName << '\n';
	out << "COMMENT:";
	QStringList commentList = fComments.split('\n');
	for (int i=0; i<commentList.size(); i++)
		out << ' ' << commentList[i] << '\n';
	out << "EDGE_WEIGHT_TYPE: EUC_2D" << '\n';
	out << "DIMENSION: " << fPoints.size() << '\n';
	out << "NODE_COORD_SECTION\n";
	for (int i=0; i<fPoints.size(); i++)
		out << i+1 << ' ' << fPoints[i].getX() << ' ' << fPoints[i].getY() << '\n';
	out << "EOF";
	out.flush();
	file.close();
	return true;
}

